"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@hibas123/utils");
const nanoid = require("nanoid/generate");
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
// TODO: Request Timeout
class Database {
    constructor(url, database, accesskey, authkey, rootkey) {
        this.initLock = new utils_1.Lock();
        this.requests = new Map();
        this.eventhandler = new Map();
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);
        u.searchParams.set("database", database);
        this.initLock.getLock().then(lock => {
            this.socket = new Database.WebSocket(u.href);
            this.socket.onopen = () => lock.release();
            this.socket.onerror = err => { console.error(err); lock.release(); };
            this.socket.onmessage = msg => {
                let message = JSON.parse(msg.data);
                switch (message.ns) {
                    case "message":
                        this.handleResponse(message.data);
                        break;
                    case "event":
                        this.handleEvent(message.data);
                        break;
                    case "error_msg":
                        // TODO: Some kind of error handling etc.
                        console.error("Server error:", message.data);
                        break;
                }
            };
        });
    }
    handleResponse({ id, error, data }) {
        let req = this.requests.get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        }
        else {
            console.error("Invalid response", id);
        }
    }
    handleEvent({ id, data }) {
        let eh = this.eventhandler.get(id);
        if (eh)
            eh(data);
    }
    async queryRequest(type, path, data, id) {
        await this.initLock.getLock().then(l => l.release());
        if (!id)
            id = nanoid(ALPHABET, 21);
        return new Promise((yes, no) => {
            this.requests.set(id, (err, data) => {
                err ? no(err) : yes(data);
                this.requests.delete(id);
            });
            this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }));
        });
    }
    async subscribe(path = [], types, options, ondata) {
        const id = nanoid(ALPHABET, 21);
        const subscriptionID = await this.queryRequest("subscribe", path, { types, options }, id);
        this.eventhandler.set(subscriptionID, ondata);
        return {
            unsubscribe: () => {
                this.queryRequest("unsubscribe", path, undefined, id);
                this.eventhandler.delete(subscriptionID);
            }
        };
    }
    collection(name) {
        return new Database.CollectionRef(this, [name]);
    }
    async collections() {
        return this.queryRequest("collections", [])
            .then(res => res.map(elm => new Database.CollectionRef(this, elm)));
    }
    close() {
        this.socket.close();
    }
}
exports.default = Database;
Database.CollectionRef = class {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
    }
    get name() {
        return this.prefix[this.prefix.length - 1];
    }
    doc(id) {
        return new Database.DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)]);
    }
    keys() {
        return this.db.queryRequest("keys", this.prefix);
    }
    async push(data) {
        let id = await this.db.queryRequest("push", this.prefix, data);
        return new Database.DocumentRef(this.db, [...this.prefix, id]);
    }
    subscribe(types, options = {}, onData) {
        return this.db.subscribe([...this.prefix], types, options, (change) => {
            onData({
                type: change.type,
                doc: {
                    data: change.data,
                    ref: new Database.DocumentRef(this.db, [...this.prefix, change.document])
                }
            });
        });
    }
    delete() {
        return this.db.queryRequest("delete-collection", this.prefix);
    }
};
Database.DocumentRef = class {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
    }
    get key() {
        return this.prefix[this.prefix.length - 1];
    }
    get() {
        return this.db.queryRequest("get", this.prefix);
    }
    set(data) {
        return this.db.queryRequest("set", this.prefix, data);
    }
    update(update) {
        throw new Error("Not implemented! WIP");
    }
    delete() {
        return this.db.queryRequest("delete", this.prefix);
    }
    subscribe(types, options = {}, onData) {
        return this.db.subscribe([...this.prefix], types, options, (change) => {
            onData({
                type: change.type,
                doc: {
                    data: change.data,
                    ref: this
                }
            });
        });
    }
    collection(name) {
        return new Database.CollectionRef(this.db, [...this.prefix, name]);
    }
};
//# sourceMappingURL=database.js.map