import { Lock } from "@hibas123/utils";
import nanoid = require("nanoid/generate");
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

type QueryTypes = "keys" | "get" | "set" | "update" | "delete" | "push" | "subscribe" | "unsubscribe" | "collections" | "delete-collection";

export type ChangeTypes = "added" | "modified" | "deleted";
type SubscribeOptions = Partial<{
   existing: boolean;
}>;

interface Change {
   data: any;
   document: string;
   type: ChangeTypes;
}

type SubscriptionCallback<T> = (change: {
   type: ChangeTypes,
   doc: {
      ref: IDocumentRef<T>,
      data: T
   }
}) => void;

export interface ICollectionRef<U = any> {
   readonly name: string;

   doc(id?: string): IDocumentRef;
   keys(): Promise<string[]>;
   push<T = U>(data: T): Promise<IDocumentRef>;

   subscribe<T = U>(types: ChangeTypes[], options: SubscribeOptions, onData: SubscriptionCallback<T>): Promise<{ unsubscribe: () => void }>;

   /**
    * Only if root key available
    */
   delete(): Promise<void>
}

export interface IDocumentRef<T = any> {
   readonly key: string;

   get(): Promise<T>;
   set(data: T): Promise<void>
   update(update: never): Promise<void>;
   delete(): Promise<void>;

   subscribe(types: ChangeTypes[], options: SubscribeOptions, onData: SubscriptionCallback<T>): Promise<{ unsubscribe: () => void }>;


   collection(name: string): ICollectionRef;
}

// TODO: Request Timeout
export default class Database {
   static WebSocket: typeof WebSocket;

   private initLock = new Lock();
   private socket: WebSocket;
   private requests = new Map<string, (err: Error, data: any) => void>();
   private eventhandler = new Map<string, (data: any) => void>();

   constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string) {
      const u = new URL(url);
      if (authkey)
         u.searchParams.set("authkey", authkey);
      if (rootkey)
         u.searchParams.set("rootkey", rootkey);
      if (accesskey)
         u.searchParams.set("accesskey", accesskey);

      u.searchParams.set("database", database);
      this.initLock.getLock().then(lock => {
         this.socket = new Database.WebSocket(u.href);

         this.socket.onopen = () => lock.release();
         this.socket.onerror = err => { console.error(err); lock.release() };
         this.socket.onmessage = msg => {
            let message: { ns: string, data: any } = JSON.parse(msg.data as string);
            switch (message.ns) {
               case "message":
                  this.handleResponse(message.data);
                  break;
               case "event":
                  this.handleEvent(message.data)
                  break;
               case "error_msg":
                  // TODO: Some kind of error handling etc.
                  console.error("Server error:", message.data);
                  break;
            }
         }
      })
   }

   private handleResponse({ id, error, data }: { id: string, error: boolean, data: string }) {
      let req = this.requests.get(id);
      if (req) {
         if (error)
            req(new Error(data), undefined);
         else
            req(undefined, data);
      } else {
         console.error("Invalid response", id);
      }
   }

   private handleEvent({ id, data }: { id: string, data: any }) {
      let eh = this.eventhandler.get(id);
      if (eh)
         eh(data);
   }

   private async queryRequest(type: QueryTypes, path: string[], data?: any, id?: string) {
      await this.initLock.getLock().then(l => l.release());
      if (!id)
         id = nanoid(ALPHABET, 21);

      return new Promise<any>((yes, no) => {
         this.requests.set(id, (err, data) => {
            err ? no(err) : yes(data)
            this.requests.delete(id);
         });
         this.socket.send(JSON.stringify({ ns: "query", data: { id, type, path, data } }))
      })
   }

   private async subscribe(path = [], types: ChangeTypes[], options: SubscribeOptions, ondata: (data: Change) => void) {
      const id = nanoid(ALPHABET, 21);
      const subscriptionID = await this.queryRequest("subscribe", path, { types, options }, id);
      this.eventhandler.set(subscriptionID, ondata);
      return {
         unsubscribe: () => {
            this.queryRequest("unsubscribe", path, undefined, id);
            this.eventhandler.delete(subscriptionID);
         }
      }
   }

   private static CollectionRef = class implements ICollectionRef {
      constructor(private db: Database, private prefix: string[]) { }

      get name() {
         return this.prefix[this.prefix.length - 1];
      }

      doc(id?: string): IDocumentRef {
         return new Database.DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)])
      }

      keys(): Promise<string[]> {
         return this.db.queryRequest("keys", this.prefix);
      }

      async push<T = any>(data: T): Promise<IDocumentRef> {
         let id = await this.db.queryRequest("push", this.prefix, data);
         return new Database.DocumentRef(this.db, [...this.prefix, id]);
      }

      subscribe<T>(types: ChangeTypes[], options: SubscribeOptions = {}, onData: SubscriptionCallback<T>) {
         return this.db.subscribe([...this.prefix], types, options, (change) => {
            onData({
               type: change.type,
               doc: {
                  data: change.data,
                  ref: new Database.DocumentRef(this.db, [...this.prefix, change.document])
               }
            })
         });
      }

      delete() {
         return this.db.queryRequest("delete-collection", this.prefix);
      }
   }

   private static DocumentRef = class <T = any> implements IDocumentRef {
      constructor(private db: Database, private prefix: string[]) { }

      get key() {
         return this.prefix[this.prefix.length - 1];
      }

      get(): Promise<T> {
         return this.db.queryRequest("get", this.prefix);
      }

      set(data: T): Promise<void> {
         return this.db.queryRequest("set", this.prefix, data);
      }

      update(update: never): Promise<void> {
         throw new Error("Not implemented! WIP");
      }

      delete(): Promise<void> {
         return this.db.queryRequest("delete", this.prefix);
      }

      subscribe(types: ChangeTypes[], options: SubscribeOptions = {}, onData: SubscriptionCallback<T>) {
         return this.db.subscribe([...this.prefix], types, options, (change) => {
            onData({
               type: change.type,
               doc: {
                  data: change.data,
                  ref: this
               }
            })
         });
      }

      collection(name: string): ICollectionRef {
         return new Database.CollectionRef(this.db, [...this.prefix, name]);
      }
   }

   public collection(name: string): ICollectionRef {
      return new Database.CollectionRef(this, [name]);
   }

   public async collections(): Promise<ICollectionRef[]> {
      return this.queryRequest("collections", [])
         .then(res => res.map(elm => new Database.CollectionRef(this, elm)));
   }

   public close() {
      this.socket.close();
   }
}